from typing import Tuple, List

from trava.fit_predictor import FitPredictConfig, FitPredictor, FitPredictorSteps
from trava.logger import TravaLogger
from trava.trava_model import TravaModel
from trava.split.result import SplitResult


class GroupAnalysisFitPredictor(FitPredictor):
    """
    If your dataset contains data related to multiple groups,
    you may want to test your model's performance on each group separately.
    See subclasses for details.
    """

    def __init__(self, group_col_name: str, steps: FitPredictorSteps = None, logger: TravaLogger = None):
        super().__init__(steps=steps or FitPredictorSteps(), logger=logger)

        self._group_col_name = group_col_name

    def _group_X_y(self, group, X, y):
        if X is None or y is None:
            return None, None

        cond = X[self._group_col_name] == group
        group_X = X[cond]
        group_y = y[cond]

        return group_X.drop(self._group_col_name, axis=1), group_y


class TrainOnOneTestOnOneFitPredictor(GroupAnalysisFitPredictor):
    """
    Trains model on each group's subset as well as gets predictions using that subset.
    """

    def _models_configs(self, raw_model, config: FitPredictConfig) -> List[Tuple[TravaModel, FitPredictConfig]]:
        split_result = config.raw_split_data
        assert split_result

        unique_groups = sorted(set(split_result.X_train[self._group_col_name].values))

        result = []
        for group in unique_groups:
            model_config = self._config_for_group(group=group, config=config)
            group_model_id = model_config.model_id + "_" + str(group)
            trava_model = TravaModel(raw_model=raw_model, model_id=group_model_id)

            result.append((trava_model, model_config))

        return result

    def _config_for_group(self, group, config: FitPredictConfig) -> FitPredictConfig:
        split_result = config.raw_split_data
        assert split_result

        group_X_train, group_y_train = self._group_X_y(group=group, X=split_result.X_train, y=split_result.y_train)
        group_X_test, group_y_test = self._group_X_y(group=group, X=split_result.X_test, y=split_result.y_test)
        group_X_valid, group_y_valid = self._group_X_y(group=group, X=split_result.X_valid, y=split_result.y_valid)

        group_split_result = SplitResult(
            X_train=group_X_train,
            y_train=group_y_train,
            X_test=group_X_test,
            y_test=group_y_test,
            X_valid=group_X_valid,
            y_valid=group_y_valid,
        )

        result = FitPredictConfig(
            raw_split_data=group_split_result,
            raw_model=config.raw_model,
            model_init_params=config.model_init_params,
            model_id=config.model_id,
            scorers_providers=config.scorers_providers,
            serializer=config.serializer,
            fit_params=config.fit_params,
            predict_params=config.predict_params,
        )

        return result


class TrainOnAllTestOnOneFitPredictor(GroupAnalysisFitPredictor):
    """
    Trains model on a whole dataset, but predictions are made for each group separately.
    """

    def __init__(self, group_col_name: str, steps: FitPredictorSteps = None, logger: TravaLogger = None):
        super().__init__(group_col_name=group_col_name, steps=steps, logger=logger)
        self._group_models: List[TravaModel] = []
        self._is_raw_model_ready = False

    def _models_configs(self, raw_model, config: FitPredictConfig) -> List[Tuple[TravaModel, FitPredictConfig]]:
        split_result = config.raw_split_data
        assert split_result

        unique_groups = sorted(set(split_result.X_train[self._group_col_name].values))

        result = []
        for group in unique_groups:
            group_config = self._config_for_group(group=group, config=config)
            group_model_id = group_config.model_id + "_" + str(group)
            result_model = TravaModel(raw_model=raw_model, model_id=group_model_id)

            result.append((result_model, group_config))

        self._group_models = [item[0] for item in result]

        return result

    def _config_for_group(self, group, config: FitPredictConfig) -> FitPredictConfig:
        split_result = config.raw_split_data
        assert split_result

        group_X_test, group_y_test = self._group_X_y(group=group, X=split_result.X_test, y=split_result.y_test)
        group_X_valid, group_y_valid = self._group_X_y(group=group, X=split_result.X_valid, y=split_result.y_valid)

        group_split_result = SplitResult(
            X_train=split_result.X_train.drop(self._group_col_name, axis=1),
            y_train=split_result.y_train,
            X_test=group_X_test,
            y_test=group_y_test,
            X_valid=group_X_valid,
            y_valid=group_y_valid,
        )

        result = FitPredictConfig(
            raw_split_data=group_split_result,
            raw_model=config.raw_model,
            model_init_params=config.model_init_params,
            model_id=config.model_id,
            scorers_providers=config.scorers_providers,
            serializer=config.serializer,
            fit_params=config.fit_params,
            predict_params=config.predict_params,
        )

        return result

    def _fit(self, trava_model: TravaModel, X, y, fit_params: dict, predict_params: dict):
        if not self._is_raw_model_ready:
            trava_model.fit(X=X, y=y, fit_params=fit_params, predict_params=predict_params)

            for group_model in self._group_models:
                if group_model != trava_model:
                    trava_model.copy(existing_model=group_model, only_fit=True)

            self._is_raw_model_ready = True



import pytest
import numpy as np
from trava.ext.sklearn.scorers import SklearnScorer


@pytest.fixture()
def metrics_kwargs():
    return {
        "param1": 1,
        "param2": "2",
    }


@pytest.fixture()
def score_func(mocker):
    score_func = mocker.Mock()
    score_func.__name__ = "test_scorer_func"
    return score_func


@pytest.mark.parametrize("needs_proba", [True, False], ids=["proba", "no_proba"])
@pytest.mark.parametrize("for_train", [True, False], ids=["train", "no_train"])
@pytest.mark.parametrize("needs_threshold", [True, False], ids=["thresh", "no_thresh"])
@pytest.mark.parametrize("greater_is_better", [True, False], ids=["greater", "no_greater"])
def test_model_data_present(
    mocker, score_func, metrics_kwargs, needs_proba, for_train, needs_threshold, greater_is_better
):
    scorer = SklearnScorer(
        score_func=score_func,
        greater_is_better=greater_is_better,
        needs_proba=needs_proba,
        needs_threshold=needs_threshold,
        **metrics_kwargs
    )
    sklearn_scorer = mocker.Mock()

    sklearn_scorer_func = mocker.patch.object(scorer, "_get_sklearn_scorer")
    sklearn_scorer_func.return_value = sklearn_scorer

    scorer_sample_weight_func = mocker.patch.object(scorer, "_sample_weight")
    sample_weight = mocker.Mock()
    scorer_sample_weight_func.return_value = sample_weight

    trava_model = mocker.Mock()
    train_raw_model = mocker.Mock()
    test_raw_model = mocker.Mock()

    def get_model(for_train: bool):
        if for_train:
            raw_model = train_raw_model
        else:
            raw_model = test_raw_model
        return raw_model

    trava_model.get_model = mocker.MagicMock(side_effect=get_model)
    X = mocker.Mock()
    X_raw = mocker.Mock()
    y = mocker.Mock()
    scorer(trava_model=trava_model, for_train=for_train, X=X, X_raw=X_raw, y=y)

    sklearn_scorer_func.assert_called_with(score_func=score_func, **metrics_kwargs)
    scorer_sample_weight_func.assert_called_with(X=X, X_raw=X_raw)
    if for_train:
        raw_model = train_raw_model
    else:
        raw_model = test_raw_model
    sklearn_scorer.assert_called_with(raw_model, X, y, sample_weight=sample_weight)


@pytest.mark.parametrize("needs_proba", [True, False], ids=["proba", "no_proba"])
@pytest.mark.parametrize("for_train", [True, False], ids=["train", "no_train"])
@pytest.mark.parametrize("needs_threshold", [True, False], ids=["thresh", "no_thresh"])
@pytest.mark.parametrize("greater_is_better", [True, False], ids=["greater", "no_greater"])
@pytest.mark.parametrize("multiclass", [True, False], ids=["multiclass", "no_multiclass"])
def test_no_model_no_data_present(
    mocker, score_func, metrics_kwargs, needs_proba, needs_threshold, greater_is_better, multiclass, for_train
):
    scorer = SklearnScorer(
        score_func=score_func,
        greater_is_better=greater_is_better,
        needs_proba=needs_proba,
        needs_threshold=needs_threshold,
        **metrics_kwargs
    )
    trava_model = mocker.Mock()
    train_y_pred = mocker.MagicMock()
    test_y_pred = mocker.MagicMock()
    train_y_pred_proba = mocker.MagicMock()
    test_y_pred_proba = mocker.MagicMock()
    if multiclass:
        train_y = [1, 2, 3]
        test_y = [1, 2, 3]
    else:
        train_y = [1, 2]
        test_y = [1, 2]

        if needs_proba:
            train_y_pred_proba = np.array([[0.1, 0.9]])
            test_y_pred_proba = np.array([[0.1, 0.9]])

    def y_pred(for_train: bool):
        if for_train:
            return train_y_pred
        else:
            return test_y_pred

    def y_pred_proba(for_train: bool):
        if for_train:
            return train_y_pred_proba
        else:
            return test_y_pred_proba

    def y(for_train: bool):
        if for_train:
            return train_y
        else:
            return test_y

    trava_model.y_pred = mocker.MagicMock(side_effect=y_pred)
    trava_model.y_pred_proba = mocker.MagicMock(side_effect=y_pred_proba)
    trava_model.y = mocker.MagicMock(side_effect=y)

    scorer(trava_model=trava_model, for_train=for_train, X=None, X_raw=None, y=None)

    if needs_proba:
        if for_train:
            y_pred_for_scoring = train_y_pred_proba if multiclass else train_y_pred_proba[:, 1]
        else:
            y_pred_for_scoring = test_y_pred_proba if multiclass else train_y_pred_proba[:, 1]
    else:
        if for_train:
            y_pred_for_scoring = train_y_pred
        else:
            y_pred_for_scoring = test_y_pred

    if for_train:
        y_true_for_scoring = train_y
    else:
        y_true_for_scoring = test_y

    score_func.assert_called_with(y_true_for_scoring, y_pred_for_scoring, **metrics_kwargs)



import inspect
import numbers


def merge_given_params_with_default(object_type, params: dict) -> dict:
    """
    Merges default params for object_type with params given by a user
    """
    all_params = dict(params)
    for key, param in inspect.signature(object_type.__init__).parameters.items():
        if not params.get(key) and param.default != param.empty:
            all_params[key] = param.default

    result = filter_params(params=all_params)
    return result


def filter_params(params: dict) -> dict:
    result = {}
    for key, value in params.items():
        if isinstance(value, _allowed_params_types()):
            result[key] = value

    return result


def _allowed_params_types() -> tuple:
    """
    Types for a model's init params that Trava is able to track.
    """
    return (
        numbers.Number,
        str,
        bool,
    )



from collections import Counter
import typing as t
import numpy as np

from trava.model_info import ModelInfo
from trava.scorer import Scorer


class GroupScorer(Scorer):
    """
    Supports group based metrics ( e.g. ranking ).
    Expects that score_func receives true scores and predicted scores.

    Init parameters
    ----------
    score_func: callable
        See superclass
    group_col_name: str
        Which column in the input data is used for storing group values
    needs_grouped_y: bool
        If True, your score_func will be called with 2d arrays instead of 1d.
        y_true will contain grouped scores for all the groups ( same goes for y_pred )
    metrics_kwargs: dict
        Any additional parameters to pass to score_func
    """

    def __init__(self, score_func: t.Callable, group_col_name: str, needs_grouped_y=False, **metrics_kwargs):
        self._group_col_name = group_col_name
        self._needs_grouped_y = needs_grouped_y

        super().__init__(
            score_func=score_func, needs_proba=False, requires_raw_model=False, requires_X_y=True, **metrics_kwargs
        )

    def _make_scorer(self, score_func: t.Callable, **metrics_kwargs) -> t.Callable:
        def scorer(model, model_info: ModelInfo, for_train: bool, X, X_raw, y):
            y_pred = model_info.y_pred(for_train=for_train)
            groups = X_raw[self._group_col_name].values
            counted_groups = list(Counter(groups).values())
            indices = np.cumsum(counted_groups)[:-1]
            grouped_y = np.split(y, indices)
            grouped_y_pred = np.split(y_pred, indices)

            if self._needs_grouped_y:
                return score_func(grouped_y, grouped_y_pred, **metrics_kwargs)
            else:
                metrics = []
                for y, y_pred in zip(grouped_y, grouped_y_pred):
                    metrics.append(score_func(y, y_pred, **metrics_kwargs))
                return np.mean(metrics)

        return scorer



from abc import ABC, abstractmethod


class ModelSerializer(ABC):
    @abstractmethod
    def load(self, path: str):
        pass

    @abstractmethod
    def save(self, model, path: str):
        pass



import pytest

from trava.trava_tracker import TravaTracker


@pytest.fixture(scope="class")
def model_id():
    return "test_model_id"


def _mocks_for_subclass_calls(mocker, tracker, model_id):
    method_names = list(_parameters_for_methods(mocker=mocker, model_id=model_id).keys())
    subclass_method_names = ["_" + method_name for method_name in method_names]

    mocks = [mocker.patch.object(tracker, method_name) for method_name in subclass_method_names]

    return mocks


def _parameters_for_methods(mocker, model_id):
    return {
        "start_tracking": dict(model_id=model_id),
        "end_tracking": dict(model_id=model_id),
        "track_set_tracking_group": dict(group="group"),
        "track_model_description": dict(model_id=model_id, description="desc"),
        "track_model_init_params": dict(model_id=model_id, params=mocker.Mock()),
        "track_fit_params": dict(model_id=model_id, params=mocker.Mock()),
        "track_predict_params": dict(model_id=model_id, params=mocker.Mock()),
        "track_metric_value": dict(model_id=model_id, name=mocker.Mock(), value=mocker.Mock()),
        "track_model_info": dict(model_id=model_id, model=mocker.Mock()),
        "track_tag": dict(model_id=model_id, tag_key=mocker.Mock(), tag_value=mocker.Mock()),
        "track": dict(model_id=model_id, privet="poka"),
    }


def test_is_enabled():
    tracker = TravaTracker(scorers=[])
    assert tracker.is_enabled
    tracker.disable()
    assert not tracker.is_enabled
    tracker.enable()
    assert tracker.is_enabled


def test_add_scorers(mocker):
    initial_scorers = [mocker.Mock()]
    tracker = TravaTracker(scorers=initial_scorers)
    new_scorers = [mocker.Mock(), mocker.Mock()]
    tracker.add_scorers(scorers=new_scorers)

    assert tracker.scorers == (initial_scorers + new_scorers)


def test_if_enabled_true(mocker, model_id):
    tracker = TravaTracker(scorers=[])

    mocks = _mocks_for_subclass_calls(mocker=mocker, tracker=tracker, model_id=model_id)
    parameters = _parameters_for_methods(mocker=mocker, model_id=model_id)

    calls = [getattr(tracker, method_name)(**params) for method_name, params in parameters.items()]

    assert len(mocks) == len(calls)
    [mock.assert_called_once() for mock in mocks]


def test_if_enabled_false(mocker, model_id):
    tracker = TravaTracker(scorers=[])
    tracker.disable()

    mocks = _mocks_for_subclass_calls(mocker=mocker, tracker=tracker, model_id=model_id)
    parameters = _parameters_for_methods(mocker=mocker, model_id=model_id)

    calls = [getattr(tracker, method_name)(**params) for method_name, params in parameters.items()]

    assert len(mocks) == len(calls)
    [mock.assert_not_called() for mock in mocks]


def test_track_model_artifact(mocker, model_id):
    serializer = mocker.Mock()
    tracker = TravaTracker(scorers=[])
    model = mocker.Mock()
    tracker.track_model_artifact(model_id=model_id, model=model, serializer=serializer)

    serializer.save.assert_called_once_with(model=model, path=mocker.ANY)



from typing import Optional

from trava.metric import Metric
from trava.model_results import ModelResult
from trava.model_serializer import ModelSerializer


class Tracker:
    """
    An interface for a tracker object.
    """

    def start_tracking(self, model_id: str, track_name: Optional[str] = None, nested: bool = False):
        pass

    def end_tracking(self, model_id: str):
        pass

    def track_set_tracking_group(self, group: str):
        pass

    def track_model_description(self, model_id: str, description: str):
        pass

    def track_model_init_params(self, model_id: str, params: dict):
        pass

    def track_fit_params(self, model_id: str, params: dict):
        pass

    def track_predict_params(self, model_id: str, params: dict):
        pass

    def track_model_results(self, model_results: ModelResult):
        pass

    def track_metric(self, model_id: str, metric: Metric, step=None):
        pass

    def track_metric_value(self, model_id: str, name: str, value, step=None):
        pass

    def track_model_info(self, model_id: str, model):
        pass

    def track_tag(self, model_id: str, tag_key: str, tag_value):
        pass

    def track_artifact(self, model_id: str, filepath):
        pass

    def track_model_artifact(self, model_id: str, model, serializer: ModelSerializer):
        pass

    def track(self, model_id: str, *args, **kwargs):
        pass

    def track_plot(self, model_id: str, fig, filename: str):
        pass



# noinspection PyPep8Naming
import pandas as pd
from sklearn.pipeline import Pipeline

from trava.fit_predictor import FitPredictConfigUpdateStep, FitPredictConfig
from trava.split.result import SplitResult
from trava.tracker import Tracker


class PreprocConfigStep(FitPredictConfigUpdateStep):
    """
    Updates split data using sklearn's Pipeline

    Init parameters
    ----------
    preprocessing: Pipeline
        Object that contains all the data transformations you need. ( without a model in it )
    """

    def __init__(self, preprocessing: Pipeline):
        self._preprocessing = preprocessing

    def fit_split_data(self, raw_split_data: SplitResult, config: FitPredictConfig, tracker: Tracker) -> SplitResult:
        X_train = self._preprocessing.fit_transform(X=raw_split_data.X_train)
        X_test = self._preprocessing.transform(X=raw_split_data.X_test)

        X_valid = raw_split_data.X_valid

        if X_valid is not None:
            X_valid = self._preprocessing.transform(X=raw_split_data.X_valid)

        result = SplitResult(
            X_train=X_train,
            X_test=X_test,
            y_train=raw_split_data.y_train,
            y_test=raw_split_data.y_test,
            X_valid=X_valid,
            y_valid=raw_split_data.y_valid,
        )

        return result

    def fit_params(
        self, fit_params: dict, fit_split_data: SplitResult, config: FitPredictConfig, tracker: Tracker
    ) -> dict:
        """
        Previous steps could already put some data in params.
        """
        return self._find_and_process_df(params=fit_params)

    def predict_params(
        self, predict_params: dict, fit_split_data: SplitResult, config: FitPredictConfig, tracker: Tracker
    ) -> dict:
        return self._find_and_process_df(params=predict_params)

    def _find_and_process_df(self, params: dict) -> dict:
        def process_df(df: pd.DataFrame):
            return self._preprocessing.transform(X=df)

        def try_process_df(maybe_df):
            if isinstance(maybe_df, pd.DataFrame):
                return process_df(df=maybe_df)

            return maybe_df

        def processed_value(value):
            if isinstance(value, (list, tuple)):
                result = []
                for list_value in value:
                    result.append(processed_value(value=list_value))

                return result
            elif isinstance(value, dict):
                for dict_key, dict_value in value.items():
                    dict_value[dict_key] = processed_value(value=dict_value)

                return value

            return try_process_df(maybe_df=value)

        for key, value in params.items():
            params[key] = processed_value(value=value)

        return params



import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="trava",
    version="0.2.8",
    description="Framework that helps to train models, compare them and track parameters&metrics along the way.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/ityutin/trava",
    author="Ilya Tyutin",
    author_email="emmarrgghh@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=find_packages(
        exclude=("*.tests", "*.tests.*", "tests.*", "tests", "*.examples", "*.examples.*", "examples.*", "examples")
    ),
    include_package_data=True,
    install_requires=["pandas", "numpy"],
)



class SplitResult:
    def __init__(self, X_train, X_test, y_train, y_test, X_valid=None, y_valid=None):
        self.X_train = X_train
        self.X_test = X_test
        self.y_train = y_train
        self.y_test = y_test
        self.X_valid = X_valid
        self.y_valid = y_valid



import numbers


class Metric:
    """
    Data class that contains result of some scorer's calculation.
    """

    def __init__(self, name: str, value, model_id: str):
        """
        Parameters
        ----------
        name: str
            What is the name of metric's function
        value:
            Value that we got after calling a scorer.
        model_id: str
            Model unique identifier, will be used for saving metrics etc
        """
        self.name = name
        self.is_scalar = isinstance(value, numbers.Number)
        self.value = value
        self.model_id = model_id



import pandas as pd
from typing import List

from trava.logger import TravaLogger
from trava.metric import Metric
from trava.model_results import ModelResult
from trava.results_handler import ResultsHandler
from trava.trava_tracker import Tracker


class PandasDfHandler(ResultsHandler):
    """
    Returns all the metrics wrapped in a pandas dataframe.
    """

    def handle(self, results: List[ModelResult], logger: TravaLogger, tracker: Tracker):
        result = pd.DataFrame()

        for model_results in results:
            test_metrics = self._metrics_dict(metrics=model_results.test_metrics(provider=self))
            other_metrics = self._metrics_dict(metrics=model_results.other_metrics(provider=self))

            all_metrics = {**test_metrics, **other_metrics}
            all_metrics["model_id"] = model_results.model_id

            result = pd.concat([result, pd.DataFrame(all_metrics, index=[0])], axis=0).reset_index(drop=True)

        return result

    @staticmethod
    def _metrics_dict(metrics: List[Metric]):
        result = {}

        for metric in metrics:
            if metric.is_scalar:
                result[metric.name] = metric.value

        return result



from typing import List

from trava.split.config import DataSplitConfig
from trava.split.result import SplitResult


# noinspection PyPep8Naming
class Splitter:
    @staticmethod
    def split(df, config: DataSplitConfig, **kwargs):
        split_logic = config.split_logic

        train_df, test_df, valid_df = split_logic.split(
            data=df, test_size=config.test_size, valid_size=config.valid_size, **kwargs
        )

        def split_X_y(source_df):
            return Splitter._split_X_y(
                df=source_df, ignore_cols=config.ignore_cols, target_col_name=config.target_col_name
            )

        # TODO: change cols to indices
        X_train, y_train = split_X_y(source_df=train_df)

        X_valid, y_eval = None, None
        if valid_df is not None:
            X_valid, y_eval = split_X_y(source_df=valid_df)

        X_test, y_test = split_X_y(source_df=test_df)

        result = SplitResult(
            X_train=X_train, X_test=X_test, y_train=y_train, y_test=y_test, X_valid=X_valid, y_valid=y_eval
        )

        for handler in config.split_result_handlers:
            result = handler.handle(split_result=result)

        return result

    @staticmethod
    def _split_X_y(df, ignore_cols: List[str], target_col_name: str):
        X, y = df.drop(ignore_cols + [target_col_name], axis=1), df[target_col_name]
        return X, y



from sklearn.metrics import auc, roc_curve

from trava.ext.results_handlers.plotter import PlotItem
from trava.ext.results_handlers.scorer_plotter import ScorerPlotter
from trava.ext.sklearn.scorers import sk_proba
from trava.metric import Metric


class ROCCurvesPlotter(ScorerPlotter):
    def plot(self, metric: Metric, fig, ax, color: str, label: str):
        ax.set_title(f"{metric.model_id}: {label} ROC")

        fpr, tpr, threshold = metric.value
        roc_auc = auc(fpr, tpr)
        ax.plot(fpr, tpr, label="{} AUC = {:0.2f}".format(metric.model_id, roc_auc), color=color)
        ax.legend(loc="lower right")
        ax.plot([0, 1], [0, 1], "r--")
        ax.set_xlim([0, 1])
        ax.set_ylim([0, 1])
        ax.set_ylabel("True Positive Rate")
        ax.set_xlabel("False Positive Rate")


class ROCCurvesPlotItem(PlotItem):
    def __init__(self):
        self.scorer = sk_proba(roc_curve)
        self.plotter = ROCCurvesPlotter()
        self.can_overlap = True



from trava.model_info import ModelInfo
from trava.scorer import OtherScorer


class FitTimeScorer(OtherScorer):
    """
    Returns the time was required to train a model.
    """

    def __init__(self):
        def fit_time(model, model_info: ModelInfo, for_train: bool, X, X_raw, y):
            return model_info.fit_time

        super().__init__(score_func=fit_time)


class PredictTimeScorer(OtherScorer):
    """
    Returns the time was required to predict an output using a model.
    """

    def __init__(self):
        def predict_time(model, model_info: ModelInfo, for_train: bool, X, X_raw, y):
            return model_info.predict_time

        super().__init__(score_func=predict_time)



from typing import List, Dict, Callable

from trava.metric import Metric
from trava.trava_model import TravaModel
from trava.scorers_provider import ScorersProvider
from trava.split.result import SplitResult


# noinspection PyPep8Naming
class Evaluator:
    """
    Calculates metrics for a provided model.

    Init parameters
    ----------
    fit_data: SplitResult
        Data that was used to train a model.
    raw_data: SplitResult
        Raw data we used to make fit_data through transformations.
    trava_model: TravaModel
        Encapsulates a real raw model and operations on it.
    """

    def __init__(self, fit_split_data: SplitResult, raw_split_data: SplitResult, trava_model: TravaModel):
        self._trava_model = trava_model
        self._fit_split_data = fit_split_data
        self._raw_split_data = raw_split_data

        self._train_metrics: Dict[str, List[Metric]] = {}
        self._test_metrics: Dict[str, List[Metric]] = {}
        self._other_metrics: Dict[str, List[Metric]] = {}

    @property
    def model_id(self) -> str:
        return self._trava_model.model_id

    @property
    def trava_model(self):
        return self._trava_model

    @property
    def fit_split_data(self) -> SplitResult:
        return self._fit_split_data

    @property
    def raw_split_data(self) -> SplitResult:
        return self._raw_split_data

    def evaluate(self, scorers_providers: List[ScorersProvider]) -> tuple:
        """
        Calculates metrics using the given scorers providers.
        Metrics will be saved in memory after evaluation.

        Parameters
        ----------
        scorers_providers: List[ScorersProvider]

        Returns
        -------
        Tuple of metrics separated by train, test and other metrics.
        """
        train_metrics = self._metrics_map(for_train=True, scorers_providers=scorers_providers, use_metric_scorers=True)

        test_metrics = self._metrics_map(for_train=False, scorers_providers=scorers_providers, use_metric_scorers=True)

        other_metrics = self._metrics_map(
            for_train=False, scorers_providers=scorers_providers, use_metric_scorers=False
        )

        self._merge_metrics(all_metrics=self._train_metrics, new_metrics=train_metrics)
        self._merge_metrics(all_metrics=self._test_metrics, new_metrics=test_metrics)
        self._merge_metrics(all_metrics=self._other_metrics, new_metrics=other_metrics)

        return train_metrics, test_metrics, other_metrics

    @staticmethod
    def _merge_metrics(all_metrics: dict, new_metrics: dict):
        for provider_id, metrics in new_metrics.items():
            existing_metrics = all_metrics.get(provider_id, [])
            existing_metrics += metrics
            all_metrics[provider_id] = existing_metrics

    def train_metrics(self, provider: ScorersProvider) -> List[Metric]:
        """
        Returns metrics calculated on train set
        """
        return self._train_metrics[provider.provider_id]

    def test_metrics(self, provider: ScorersProvider) -> List[Metric]:
        """
        Returns metrics calculated on test set
        """
        return self._test_metrics[provider.provider_id]

    def other_metrics(self, provider: ScorersProvider) -> List[Metric]:
        """
        Returns custom train/test-independent metrics
        """
        return self._other_metrics[provider.provider_id]

    @staticmethod
    def _metrics_scorers_getter() -> Callable[[ScorersProvider], list]:
        return lambda provider: provider.metric_scorers()

    @staticmethod
    def _other_scorers_getter() -> Callable[[ScorersProvider], list]:
        return lambda provider: provider.other_scorers()

    def _metrics_map(
        self, for_train: bool, scorers_providers: List[ScorersProvider], use_metric_scorers: bool
    ) -> Dict[str, List[Metric]]:
        """
        Calculates metrics using the saved scorers providers
        """
        result = {}
        metrics_cache: Dict[str, Metric] = {}

        if self._fit_split_data is None:
            X = None
            X_raw = None
            y = None
        elif for_train:
            X = self._fit_split_data.X_train
            X_raw = self._raw_split_data.X_train
            y = self._fit_split_data.y_train
        else:
            X = self._fit_split_data.X_test
            X_raw = self._raw_split_data.X_test
            y = self._fit_split_data.y_test

        for provider in scorers_providers:
            if use_metric_scorers:
                scorers = provider.metric_scorers()
            else:
                scorers = provider.other_scorers()

            provider_metrics = []
            for scorer in scorers:
                if metrics_cache.get(scorer.func_name):
                    provider_metrics.append(metrics_cache[scorer.func_name])
                    continue

                metric = self._get_metric(X=X, X_raw=X_raw, y=y, for_train=for_train, scorer=scorer)
                metrics_cache[scorer.func_name] = metric

                provider_metrics.append(metric)

            result[provider.provider_id] = provider_metrics

        return result

    def _get_metric(self, for_train: bool, X, X_raw, y, scorer) -> Metric:
        """
        Calculate metric using the given scorer
        """
        value = scorer(trava_model=self.trava_model, for_train=for_train, X=X, X_raw=X_raw, y=y)
        return Metric(name=scorer.func_name, value=value, model_id=self.model_id)

    def unload_data(self):
        """
        If storing the data is unwanted behavior, you can unload it.
        """
        del self._fit_split_data
        del self._raw_split_data

        self._fit_split_data = None
        self._raw_split_data = None



from typing import Optional

from trava.ext.split_logic.basic import BasicSplitLogic
from trava.split.logic import SplitLogic


class TimeSplitLogic(SplitLogic):
    """
    Basic time split. It works just by sorting the data
    by time_col and taking the tail of the data.
    Assumes that you work with pandas data frame.

    Init parameters
    ----------
    time_col: str
        What is the column that contains time data.
    need_to_sort:
        Whether your data is already sorted
    group_col:
        Whether your data contains groups and you want
        to split every group separately.
    """

    def __init__(self, time_col: str, need_to_sort=True, group_col: Optional[str] = None):
        self._time_col = time_col
        self._need_to_sort = need_to_sort
        self._split_logic = BasicSplitLogic(shuffle=False, validation_from_test=True, group_col=group_col)

    def split(self, data, test_size: float, valid_size: float, **kwargs) -> tuple:
        sorted_data = data

        if self._need_to_sort:
            sorted_data = data.sort_values(self._time_col)

        return self._split_logic.split(data=sorted_data, test_size=test_size, valid_size=valid_size)



from typing import Optional
from sklearn.model_selection import train_test_split

from trava.split.logic import SplitLogic


class BasicSplitLogic(SplitLogic):
    """
    Basically just a wrapper around sklearn's train_test_split.
    Can be used in combination with another split logic.
    """

    def __init__(
        self,
        random_state: int = 15,
        shuffle: bool = True,
        validation_from_test: bool = True,
        group_col: Optional[str] = None,
        stratify: bool = None,
    ):
        self._random_state = random_state
        self._shuffle = shuffle
        self._stratify = stratify
        self._valid_from_test = validation_from_test
        self._group_col = group_col

    def split(self, data, test_size: float, valid_size: float = 0.0, **kwargs) -> tuple:
        data_to_split = data

        if self._group_col:
            unique_groups = data[self._group_col].unique()
            data_to_split = unique_groups

        train_data, test_data = train_test_split(
            data_to_split,
            test_size=test_size,
            shuffle=self._shuffle,
            stratify=self._stratify,
            random_state=self._random_state,
            **kwargs
        )

        valid_data = None
        if valid_size > 0.0:
            if self._valid_from_test:
                data_for_valid_split = test_data
            else:
                data_for_valid_split = train_data
            data_left, valid_data = train_test_split(
                data_for_valid_split,
                test_size=valid_size,
                shuffle=self._shuffle,
                stratify=self._stratify,
                random_state=self._random_state,
                **kwargs
            )
            if self._valid_from_test:
                test_data = data_left
            else:
                train_data = data_left

        if self._group_col:
            train_data = data[data[self._group_col].isin(train_data)]
            test_data = data[data[self._group_col].isin(test_data)]
            if valid_data is not None:
                valid_data = data[data[self._group_col].isin(valid_data)]

        return train_data, test_data, valid_data



from time import time  # type: ignore
from typing import Optional, Any, Dict

import numpy as np
import pandas as pd

from trava.model_info import ModelInfo


class _RawModelImitation:
    """
    If TravaModel doesn't have raw model in memory anymore,
    it will be replaced with an object of this class to be able
    to provide results made by the real model.

    Init parameters
    ----------
    y_pred: np.array
        hard labels, predicted by a raw model
    y_pred_proba: np.array
        classes' probabilities, predicted by a raw model
    """

    def __init__(self, y_pred, y_pred_proba: Optional[Any] = None):
        self._y_pred = y_pred
        self._y_pred_proba = y_pred_proba

    def predict(self, X):
        return self._y_pred

    def predict_proba(self, X):
        assert self._y_pred_proba is not None, "You haven't provided y_pred_proba"
        return self._y_pred_proba


class _RawModelWrapper:
    def __init__(self, raw_model, y_pred: np.ndarray, y_pred_proba: np.ndarray):
        self._raw_model = raw_model

        self._y_pred: np.ndarray = y_pred
        self._y_pred_proba: np.ndarray = y_pred_proba

    def __eq__(self, o: object) -> bool:
        return self._raw_model == o

    def __getattr__(self, attr):
        return getattr(self._raw_model, attr)

    def predict(self, *args, **kwargs):
        return self._y_pred

    def predict_proba(self, *args, **kwargs):
        return self._y_pred_proba


class TravaModel(ModelInfo):
    """
    Wrapper of any model you want to use with Trava.
    Stores both true and predicted labels as well as some metadata.

    Init parameters
    ----------
    raw_model: sklearn-style model
        Model that supports fit, predict and predict_proba methods.
    model_id: str
        Model unique identifier, will be used for saving metrics etc
    """

    def __init__(self, raw_model, model_id: str):
        self._raw_model = raw_model
        self._needs_proba = hasattr(raw_model, "predict_proba") and callable(raw_model.predict_proba)
        self._model_id = model_id

        self._y_train = None
        self._y_train_pred = None
        self._y_train_pred_proba = None
        self._y_test = None
        self._y_test_pred = None
        self._y_test_pred_proba = None

        self._fit_params: Dict[str, Any] = {}
        self._predict_params: Dict[str, Any] = {}
        self._fit_time: Optional[float] = None
        self._predict_time: Optional[float] = None

    def copy(
        self, model_id: Optional[str] = None, existing_model: Optional["TravaModel"] = None, only_fit: bool = True
    ):
        """
        Makes a copy of the trava_model, but with another model_id
        """
        if (not model_id and not existing_model) or (model_id and existing_model):
            raise Exception("Pass either model_id or existing_model")

        result = existing_model
        if not result:
            # TODO: ugly
            assert model_id
            result = TravaModel(raw_model=self.raw_model, model_id=model_id)

        result._y_train = np.copy(self._y_train)
        result._y_train_pred = np.copy(self._y_train_pred)
        result._y_train_pred_proba = np.copy(self._y_train_pred_proba)
        if not only_fit:
            result._y_test = np.copy(self._y_test)
            result._y_test_pred = np.copy(self._y_test_pred)
            result._y_test_pred_proba = np.copy(self._y_test_pred_proba)

        if self.fit_params:
            result._fit_params = self.fit_params.copy()
        if not only_fit and self.predict_params:
            result._predict_params = self.predict_params.copy()
        result._fit_time = self.fit_time
        if not only_fit:
            result._predict_time = self.predict_time

        return result

    @property
    def model_id(self) -> str:
        return self._model_id

    def get_model(self, for_train: bool):
        if for_train:
            y_pred = self._y_train_pred
            y_pred_proba = self._y_train_pred_proba
        else:
            y_pred = self._y_test_pred
            y_pred_proba = self._y_test_pred_proba

        if self._raw_model:
            return _RawModelWrapper(raw_model=self._raw_model, y_pred=y_pred, y_pred_proba=y_pred_proba)

        if y_pred is None:
            raise ValueError("y_pred is missing as well as raw model, unexpected behaviour")

        return _RawModelImitation(y_pred=y_pred, y_pred_proba=y_pred_proba)

    @property
    def raw_model(self):
        return self._raw_model

    @property
    def fit_params(self) -> dict:
        return self._fit_params

    @property
    def fit_time(self) -> Optional[float]:
        return self._fit_time

    @property
    def predict_time(self) -> Optional[float]:
        return self._predict_time

    @property
    def predict_params(self) -> dict:
        return self._predict_params

    @property
    def is_classification_model(self) -> bool:
        return self._needs_proba

    def y(self, for_train: bool) -> np.array:
        if for_train:
            return self._y_train
        else:
            return self._y_test

    def y_pred(self, for_train: bool) -> np.array:
        return self._y_train_pred if for_train else self._y_test_pred

    def y_pred_proba(self, for_train: bool) -> np.array:
        return self._y_train_pred_proba if for_train else self._y_test_pred_proba

    def fit(self, X, y, fit_params: Optional[dict] = None, predict_params: Optional[dict] = None):
        """
        Fits the saved model

        Parameters
        ----------
        X: pandas dataframe
            Features for the model
        y: numpy array, pandas series
            Target for the model
        fit_params: dict
            Custom params to use when calling model's fit method
        predict_params: dict
                Custom params to use when calling model's predict method

        """
        self._fit_params = fit_params or {}
        self._predict_params = predict_params or {}

        start = time()
        self._raw_model.fit(X, y, **self._fit_params)
        end = time()
        self._fit_time = end - start

        self._y_train = y.values if isinstance(y, pd.Series) else y
        self._y_train_pred = self._raw_model.predict(X, **self._predict_params)

        if self._needs_proba:
            self._y_train_pred_proba = self._raw_model.predict_proba(X, **self._predict_params)

    def predict(self, X, y):
        """
        Calls predict and (if possible) predict proba on the saved model

        Parameters
        ----------
        X: pandas dataframe
            Features for the model
        """
        start = time()
        self._y_test = y.values if isinstance(y, pd.Series) else y
        predict_params = self._predict_params or {}
        self._y_test_pred = self._raw_model.predict(X, **predict_params)
        end = time()
        self._predict_time = end - start

        if self._needs_proba:
            self._y_test_pred_proba = self._raw_model.predict_proba(X, **predict_params)

    def unload_model(self):
        """
        Deletes fitted model from memory
        """
        del self._raw_model
        self._raw_model = None



from typing import Optional, Dict, Any

import pytest

from tests.objects_for_tests import TestScorer, TestAnyScorer
from trava.model_info import ModelInfo
from trava.scorer import Scorer


@pytest.fixture
def score_func(mocker, score_func_name):
    result = mocker.Mock()
    result.__name__ = score_func_name

    return result


@pytest.fixture(scope="class")
def score_func_name():
    return "test_score_func"


def test_is_any_scorer(mocker, score_func):
    scorer = TestScorer(score_func=score_func)
    assert not scorer.is_other_scorer
    any_scorer = TestAnyScorer(model_func=score_func)
    assert any_scorer.is_other_scorer


def test_func_name(mocker, score_func, score_func_name):
    scorer = TestScorer(score_func=score_func)
    assert scorer._func_name == score_func_name


def test_require_raw_model(mocker, score_func):
    scorer = TestScorer(score_func=score_func, requires_raw_model=True)
    trava_model = mocker.Mock()
    trava_model.raw_model.return_value = None

    X = mocker.MagicMock()
    X_raw = mocker.MagicMock()
    y = mocker.MagicMock()

    with pytest.raises(Exception):
        scorer(trava_model=trava_model, for_train=True, X=X, X_raw=X_raw, y=y)

    trava_model.raw_model.return_value = mocker.MagicMock()

    scorer(trava_model=trava_model, for_train=True, X=X, X_raw=X_raw, y=y)


def test_require_X_y(mocker, score_func):
    scorer = TestScorer(score_func=score_func, requires_X_y=True)
    trava_model = mocker.Mock()
    trava_model.raw_model.return_value = mocker.MagicMock()

    X = None
    X_raw = mocker.MagicMock()
    y = mocker.MagicMock()

    with pytest.raises(Exception):
        scorer(trava_model=trava_model, for_train=True, X=X, X_raw=X_raw, y=y)

    X = mocker.MagicMock()
    X_raw = None
    y = mocker.MagicMock()

    with pytest.raises(Exception):
        scorer(trava_model=trava_model, for_train=True, X=X, X_raw=X_raw, y=y)

    X = mocker.MagicMock()
    X_raw = mocker.MagicMock()
    y = None

    with pytest.raises(Exception):
        scorer(trava_model=trava_model, for_train=True, X=X, X_raw=X_raw, y=y)

    X = mocker.MagicMock()
    X_raw = mocker.MagicMock()
    y = mocker.MagicMock()
    scorer(trava_model=trava_model, for_train=True, X=X, X_raw=X_raw, y=y)


@pytest.mark.parametrize("for_train", [True, False])
def test_scorer_func_input(mocker, score_func, for_train):
    kwargs = {"a": 3, "b": "3fff"}
    scorer = TestScorer(score_func=score_func, **kwargs)
    internal_scorer = mocker.patch.object(scorer, "_scorer")

    trava_model = mocker.Mock()
    trava_model.raw_model.return_value = mocker.MagicMock()
    raw_model = mocker.Mock()
    trava_model.get_model.return_value = raw_model

    X = mocker.MagicMock()
    X_raw = mocker.MagicMock()
    y = mocker.MagicMock()

    scorer(trava_model=trava_model, for_train=for_train, X=X, X_raw=X_raw, y=y, **kwargs)

    internal_scorer.assert_called_once_with(
        model=raw_model, model_info=trava_model, for_train=for_train, X=X, X_raw=X_raw, y=y, **kwargs
    )


def test_make_scorer(mocker, score_func):
    class LocalTestScorer(Scorer):
        def __init__(
            self,
            score_func: callable,
            needs_proba=False,
            requires_raw_model=False,
            requires_X_y=False,
            name: Optional[str] = None,
            **metrics_kwargs
        ):
            self._make_scorer_score_func = None
            self._make_scorer_metrics_kwargs: Dict[str, Any] = {}
            super().__init__(score_func, needs_proba, requires_raw_model, requires_X_y, name, **metrics_kwargs)

        def _make_scorer(self, score_func: callable, **metrics_kwargs) -> callable:
            self._make_scorer_score_func = score_func
            self._make_scorer_metrics_kwargs = metrics_kwargs

            def scorer(model, model_info: ModelInfo, for_train: bool, X, X_raw, y):
                return 1

            return scorer

    kwargs = {"a": 3, "b": "3fff"}
    scorer = LocalTestScorer(score_func=score_func, **kwargs)

    assert scorer._make_scorer_score_func == score_func
    assert scorer._make_scorer_metrics_kwargs == kwargs
